# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Data provider sub-package."""

# flake8: noqa: F401
from .azure_data import AzureData
from .sentinel_core import MicrosoftSentinel
